import json
import copy
import traceback as tb
from datetime import datetime as dt
from multiprocessing import Pool
from tabulate import tabulate
from typing import Optional
import rapidfuzz as rf

import pandas as pd
import psutil
from pydantic import BaseModel, Field, ValidationError, conint, conlist, validator

from ipc.utility.helper import construct_args_for_pool, timeit, list_to_set_in_dict, lower_and_strip
from settings import BUCKET, REPOS, ROOT_DIR

pd.set_option('display.max_rows', None)
pd.set_option('display.max_columns', None)
pd.set_option('display.width', 1000)
pd.set_option('display.colheader_justify', 'center')
pd.set_option('display.precision', 3)

# Load json file containing valid ontology valyes
with open(f'{ROOT_DIR}/knowledge_base/valid_names.json') as file:
    valid_names = json.load(file)

# Creating valid name but lowered and stripped
valid_names_lowered = copy.deepcopy(valid_names)
valid_names_lowered = lower_and_strip(valid_names_lowered)

# Convert all lists to set for faster lookups
valid_names = list_to_set_in_dict(valid_names)
valid_names_lowered = list_to_set_in_dict(valid_names_lowered)


class ValidateMetadata(BaseModel):
    """
    Validation model to check the validity of dataset level metadata
    """

    # Schema Checks [SC]
    index: dict = Field(alias='__index__')  # Using an alias since '__index__' is causing a problem
    dataset_id: str
    dataset_source: str
    description: str
    publication: str
    organism: conlist(str, min_items=1)
    curation_version: Optional[str]
    total_num_samples: conint(gt=0)
    year: str
    tissue: conlist(str, min_items=1)
    donor_age: Optional[str]
    disease: conlist(str, min_items=1)
    kw_cell_type: conlist(str, min_items=1)
    kw_donor_sex: Optional[str]
    kw_disease_subtype: Optional[str]
    kw_sample_type: Optional[str]
    kw_drug: conlist(str, min_items=1)
    kw_cell_line: conlist(str, min_items=1)
    kw_data_type: str

    """-------Field Specific Checks [FC]-------"""

    @validator('index')
    def check_index(cls, val):
        """
        Required checks for index field
        Args:
            val:

        Returns:
            val
        """
        if 'data_required' not in val:
            raise ValueError('"data_required" field not present in "__index__"')
        else:
            if val['data_required'] not in {'true', 'false'}:
                raise ValueError('"data_required" can only contain the values "true" and "false".')
        return val

    @validator('dataset_id')
    def check_dataset_id(cls, val):
        """
        Check that there should not be any spaces in dateset id value
        Args:
            val: value of the field

        Returns:
            val
        """
        if ' ' in val:
            raise ValueError(f'Value contains one or more spaces. | Erroneous Value: {val}')
        return val

    @validator('year')
    def check_year(cls, val):
        """
        Year has to be an integer and not greater than current year
        Args:
            val:

        Returns:
            val
        """
        is_int = True
        try:
            # converting to integer
            int(val)
        except ValueError:
            is_int = False
        if not is_int:
            raise ValueError(f'Value not a valid value for "year". | Erroneous Value: {val}')
        if is_int:
            if int(val) > dt.now().year or int(val) < 1900:
                raise ValueError(
                    f'Value is invalid since it cannot be greater than current year or less than 1900.'
                    f' | Erroneous Value: {val}')
        return val

    @validator('dataset_source')
    def check_valid_dataset_source(cls, val, field):
        """
        'dataset_source' can only be one of the predefined values.
        Args:
            val:
            field: name of field currently being validated i.e. dataset_source

        Returns:
            val
        """
        field_to_check = field.name
        if field_to_check[0] == '_':
            field_to_check = field_to_check[1:]
        valid = valid_names[field_to_check]
        if val not in valid:
            raise ValueError(f'Type 2.3 | Invalid value | Erroneous Value: {val}')
        return val

    """-------Ontology Value Checks [OC]-------"""

    @validator('kw_data_type', 'disease', 'kw_cell_type', 'kw_cell_line', 'kw_drug', 'tissue',
               allow_reuse=True)
    def check_for_default_values(cls, val, field):
        """
        Check for a single None value
        Args:
            val:
            field:

        Returns:

        """
        field_to_check = field.name
        if field_to_check[0] == '_':
            field_to_check = field_to_check[1:]
        if field_to_check == 'disease':
            if 'Normal' in val and len(val) > 1:
                raise ValueError(
                    f'"{field_to_check}" contains one or more default value "Normal" | Erroneous Value: {val}')
        elif 'None' in val or 'Normal' in val:
            if len(val) > 1:
                raise ValueError(f'"{field_to_check}" contains one or more default values. | Erroneous Value: {val}')
        if 'none' in val or 'NONE' in val:
            raise ValueError(f'"{field_to_check}" contains invalid value. | Erroneous Value: {val}')
        return val

    @validator('kw_data_type', 'disease', 'kw_cell_type', 'kw_cell_line', 'kw_drug', 'tissue',
               allow_reuse=True,
               each_item=True)
    def check_valid_names(cls, val, field):
        """
        Check that the values of the above field case-sensitively match the valid values as in the ontology knowledge
        base
        Args:
            val:
            field:

        Returns:
            val
        """
        field_to_check = field.name
        # field.name at times comes with a leading underscore
        if field_to_check[0] == '_':
            field_to_check = field_to_check[1:]
        if field_to_check == 'kw_cell_line':
            # Specifically for kw_cell_line, valid names are stores separately.
            valid = valid_names[field_to_check]['valid_cell_line_names']
            valid_lowered_stripped = valid_names_lowered[field_to_check]['valid_cell_line_names']
        else:
            valid = valid_names[field_to_check]
            valid_lowered_stripped = valid_names_lowered[field_to_check]
        if val != 'None' and val != 'none' and val != 'Normal':
            if val not in valid:
                if val.lower().strip() in valid_lowered_stripped:
                    possible_match = rf.process.extract(val, valid, scorer=rf.fuzz.ratio, limit=1)[0][0]
                    raise ValueError(f'Valid Value BUT with case/whitespaces mismatch '
                                     f'| Erroneous Value: "{val}" | Possible match: "{possible_match}"')
                else:
                    raise ValueError(f'Type 2.4 | Invalid Value | Erroneous Value: "{val}"')
        return val

    """-------Logical Checks [LC]--------"""

    @validator('kw_cell_line',
               each_item=True)
    def check_kw_cell_line_organism_agreement(cls, val, values, field):
        """
        Check that the value of cell line is valid for the particular organism
        Args:
            val: value of the field
            values: dict of values of the model object
            field: field being validated

        Returns:

        """
        field_to_check = field.name
        if field_to_check[0] == '_':
            field_to_check = field_to_check[1:]
        if val != 'None' and val != 'none':
            if 'organism' in values:
                organism_list = values.get('organism')
                organisms_to_check_for_cell_line = {org: cell_lines for org, cell_lines in
                                                    valid_names[field_to_check]['organism'].items() if
                                                    org in organism_list}
                if organisms_to_check_for_cell_line:
                    if not any([True for v in organisms_to_check_for_cell_line.values() if val in v]):
                        raise ValueError(
                            f'Not a valid cell line for organism = {list(organisms_to_check_for_cell_line.keys())}'
                            f' | Erroneous Value: "{val}"')
        return val

    @validator('kw_cell_line',
               each_item=True)
    def check_kw_cell_line_disease_agreement(cls, val, values, field):
        """
        Check that the value of cell line is valid for the particular disease
        Args:
            val: value of the field
            values: dict of values of the model object
            field: field being validated

        Returns:

               """
        field_to_check = field.name
        if field_to_check[0] == '_':
            field_to_check = field_to_check[1:]
        if val != 'None' and val != 'none':
            if 'disease' in values:
                disease_list_from_input = values.get('disease')
                diseases_to_check_for_cell_line = {disease: cell_lines for disease, cell_lines in
                                                   valid_names[field_to_check]['disease'].items() if
                                                   disease in disease_list_from_input}
                if diseases_to_check_for_cell_line:
                    if not any([True for v in diseases_to_check_for_cell_line.values() if val in v]):
                        raise ValueError(
                            f'Not a valid cell line for disease = {list(diseases_to_check_for_cell_line.keys())}'
                            f' | Erroneous Value: "{val}"')
        return val


@timeit
def collect_error_df():
    """
    Runs validation checks on the repos mentioned in the config filr to adhere to the model defined in
    dataset_metadata_validator.py
    Returns:
        Dataframe containing errors and details
    """
    try:
        compiled_error_df = pd.DataFrame(columns=['Repo', 'Dataset ID', 'Field', 'Error Message'])
        print(f'Constructing arguments list for process pool...')
        num_repos = len(REPOS)
        args_for_pool = construct_args_for_pool(REPOS, BUCKET)
        all_error_dfs = []
        ctr = 1
        print(f'\nWill use {psutil.cpu_count()} cores at my disposal...\n')
        for repo in REPOS:
            tstart = dt.now()
            print(f'Working on repo: {repo} | Count: {ctr}/{num_repos}')
            process_pool = Pool(processes=psutil.cpu_count(logical=False))
            error_df_results = process_pool.starmap(collect_errors_per_dataset, args_for_pool[repo])
            print(f'Time Taken: {dt.now() - tstart}\n')
            all_error_dfs.extend(error_df_results)
            ctr += 1
        compiled_error_df = pd.concat([compiled_error_df, pd.concat(all_error_dfs)], ignore_index=True)
        compiled_error_df['Field'] = compiled_error_df['Field'].apply(lambda row: row[0])
        return compiled_error_df
    except Exception:
        print(f'Error when collecting errors: {tb.format_exc()}')
        return pd.DataFrame()


def collect_errors_per_dataset(repo, key, entry_meta):
    try:
        ValidateMetadata(**entry_meta)
    except ValidationError as e:
        error_df = pd.DataFrame(e.errors())
        error_df.rename(columns={'loc': 'Field',
                                 'msg': 'Error Message',
                                 'type': 'Error Type'}, inplace=True)
        error_df['Dataset ID'] = entry_meta['dataset_id'] if 'dataset_id' in entry_meta else None
        error_df['Repo'] = repo
        error_df['Key'] = key
        return error_df
    return pd.DataFrame(columns=['Repo', 'Dataset ID', 'Field', 'Error Message'])


def check_metadata_for_errors(metadata, print_table=False):
    """
    Utility to return errors for a single dataset level metadata dict
    Args:
        metadata: dict containing the dataset level metadata to be checked
        print_table: flag to print the results as a table or not

    Returns:
        errors and desc :  errors as an error dictif
        status (boolean): Status for if the given metadata passed all checks or not
    """
    try:
        ValidateMetadata(**metadata)
    except ValidationError as err:
        error_df = pd.DataFrame(err.errors())
        error_df.rename(columns={'loc': 'Field',
                                 'msg': 'Error Message',
                                 'type': 'Error Type'}, inplace=True)
        error_df['Dataset ID'] = metadata['dataset_id'] if 'dataset_id' in metadata else None
        error_df['Field'] = error_df['Field'].apply(lambda row: row[0])
        error_df['Error Message'] = error_df['Error Message'].str.wrap(20)
        if print_table:
            # 'print_table' is True: returns formatted table for the command line
            print(tabulate(error_df, headers="keys", tablefmt="fancy_grid"))
        return error_df.to_dict('records'), False
    return {}, True


# if __name__ == '__main__':
#     with open(r'/home/elucidata/Downloads/Metadata Test Samples/test_dataset_metadata_1.json', 'r') as fp:
#         test_sample = json.load(fp)
#     errors, status = check_metadata_for_errors(test_sample, print_table=True)
#     print(errors)
#     print(status)
